local ffi = require('ffi')
local band, bor, bnot, lshift
do
  local _obj_0 = require('bit')
  band, bor, bnot, lshift = _obj_0.band, _obj_0.bor, _obj_0.bnot, _obj_0.lshift
end
local CIA_cdef = [[	struct __attribute__((__packed__)) {
		uint8_t porta, portb;
		uint8_t ddra, ddrb;
		uint16_t ta, tb;
		uint32_t tod;
		uint8_t sdr;
		uint8_t icr;
		uint8_t cra, crb;
	}
]]
local CIA
CIA = function(system, baseAddr, readReg, writeReg)
  local cpu = system:getCPU()
  local regs = nil
  local timerAStart, timerBStart = 0xffff, 0xffff
  return ffi.metatype(ffi.typeof(CIA_cdef), {
    __new = function(self)
      local cia = ffi.new(self)
      regs = ffi.cast('char*', cia)
      cia.ta, cia.tb = timerAStart, timerBStart
      system:addDevice(baseAddr, baseAddr + 0xff, cia)
      return cia
    end,
    __index = {
      PORTA = 0x00,
      PORTB = 0x01,
      DDRA = 0x02,
      DDRB = 0x03,
      TALO = 0x04,
      TAHI = 0x05,
      TBLO = 0x06,
      TBHI = 0x07,
      TODTS = 0x08,
      TODS = 0x09,
      TODM = 0x0a,
      TODH = 0x0b,
      SDR = 0x0c,
      ICR = 0x0d,
      CRA = 0x0e,
      CRB = 0x0f,
      read = function(self, addr)
        local reg = band(addr, 0xff0f) - baseAddr
        return readReg(reg)
      end,
      write = function(self, addr, val)
        local reg = band(addr, 0xff0f) - baseAddr
        return writeReg(reg, val)
      end,
      _readReg = function(self, reg)
        return regs[reg]
      end,
      _writeReg = function(self, reg, val)
        local _exp_0 = reg
        if 0x00 == _exp_0 then
          val = bor(band(val, self.ddra), band(self.porta, band(bnot(self.ddra), 0xff)))
        elseif 0x01 == _exp_0 then
          val = bor(band(val, self.ddrb), band(self.portb, band(bnot(self.ddrb), 0xff)))
        elseif 0x04 == _exp_0 then
          timerAStart = val + band(timerAStart, 0xff00)
        elseif 0x05 == _exp_0 then
          timerAStart = lshift(val, 8) + band(timerAStart, 0xff)
        elseif 0x06 == _exp_0 then
          timerBStart = val + band(timerBStart, 0xff00)
        elseif 0x07 == _exp_0 then
          timerBStart = lshift(val, 8) + band(timerBStart, 0xff)
        elseif 0x0e == _exp_0 then
          if band(val, 0x10) == 0x10 then
            self.ta = timerAStart
          end
        elseif 0x0f == _exp_0 then
          if band(val, 0x10) == 0x10 then
            self.tb = timerBStart
          end
        end
        regs[reg] = val
        return val
      end,
      _cnt = function(self)
        if band(self.cra, 0x21) == 0x21 then
          self:_updateTimerA()
        end
        if band(self.crb, 0x21) == 0x21 then
          return self:_updateTimerB()
        end
      end,
      _tick = function(self)
        if band(self.cra, 0x21) == 0x01 then
          self:_updateTimerA()
        end
        if band(self.crb, 0x21) == 0x01 then
          return self:_updateTimerB()
        end
      end,
      _updateTimerA = function(self)
        local time = self.ta - 1
        if time == -1 then
          self.cra = bor(self.cra, 0x01)
          if band(self.cra, 0x08) == 0 then
            self.ta = timerAStart
          else
            self.ta = 0
            self.cra = band(self.cra, 0xfe)
          end
          if band(self.icr, 0x01) then
            return cpu:irq()
          end
        else
          self.ta = time
          self.icr = band(self.icr, 0xfe)
        end
      end,
      _updateTimerB = function(self)
        local time = self.tb - 1
        if time == -1 then
          self.icr = bor(self.icr, 0x02)
          if band(self.crb, 0x08) == 0 then
            self.tb = timerBStart
          else
            self.tb = 0
            self.crb = band(self.crb, 0xfe)
          end
          if band(self.icr, 0x02) then
            return cpu:irq()
          end
        else
          self.tb = time
          self.icr = band(self.icr, 0xfd)
        end
      end
    }
  })()
end
return function(...)
  return CIA(...)
end
