local ffi = require('ffi')
local bor, band, bxor, bnot, lshift, rshift
do
  local _obj_0 = require('bit')
  bor, band, bxor, bnot, lshift, rshift = _obj_0.bor, _obj_0.band, _obj_0.bxor, _obj_0.bnot, _obj_0.lshift, _obj_0.rshift
end
local min, max
do
  local _obj_0 = math
  min, max = _obj_0.min, _obj_0.max
end
local OPCODES, MNEMONICS = require('CPU.opcodes'), require('CPU.mnemonics')
local STACK = 0x100
local NMI = 0xfffa
local RESET = 0xfffc
local IRQ = 0xfffe
local CPU_cdef = [[	struct {
		uint8_t A, X, Y, SP, P;
		uint16_t PC;

		uint8_t  _port[2];
		uint16_t _ir_pending;
		uint8_t  _cycles;
		uint64_t _instCounter;
	}
]]
local formatStatus
formatStatus = function(s)
  return ("%s%s%s%s%s%s%s%s"):format((band(s, 0x80) > 0 and 'N' or '.'), (band(s, 0x40) > 0 and 'V' or '.'), '-', (band(s, 0x10) > 0 and 'B' or '.'), (band(s, 0x08) > 0 and 'D' or '.'), (band(s, 0x04) > 0 and 'I' or '.'), (band(s, 0x02) > 0 and 'Z' or '.'), (band(s, 0x01) > 0 and 'C' or '.'))
end
local printState
printState = function(self, op, size, mem)
  local ops
  do
    local _accum_0 = { }
    local _len_0 = 1
    for i = 0, size - 1 do
      _accum_0[_len_0] = ("%02x"):format(mem:_readRaw(self.PC + i))
      _len_0 = _len_0 + 1
    end
    ops = _accum_0
  end
  return io.write(("%9d %s PC: %04x OP: %6s A: %02x X: %02x Y: %02x S: %02x %s\n"):format(tonumber(self._instCounter), MNEMONICS[op], self.PC, table.concat(ops), self.A, self.X, self.Y, self.SP, formatStatus(self.P)))
end
local CPU
CPU = function(mem, startAddr)
  return ffi.metatype(ffi.typeof(CPU_cdef), {
    __new = function(self)
      local cpu = ffi.new(self)
      cpu:reset(startAddr)
      return cpu
    end,
    __index = {
      step = function(self)
        if self._cycles == 0 then
          if self._ir_pending > 0 then
            if self._ir_pending == NMI or not self:testStatus(0x04) then
              self:_interrupt()
            end
          end
          self:_nextInst()
        end
        self._cycles = self._cycles - 1
      end,
      readPort = function(self, addr)
        return self._port[addr]
      end,
      writePort = function(self, addr, val)
        self._port[addr] = val
        return val
      end,
      read8 = function(self, addr)
        return mem:read8(addr)
      end,
      write8 = function(self, addr, val)
        return mem:write8(addr, val)
      end,
      read16 = function(self, addr)
        return mem:read16(addr)
      end,
      write16 = function(self, addr, val)
        return mem:write16(addr, val)
      end,
      push8 = function(self, val)
        self:write8(STACK + self.SP, val)
        self.SP = self.SP - 1
      end,
      pop8 = function(self)
        self.SP = self.SP + 1
        return self:read8(STACK + self.SP)
      end,
      push16 = function(self, val)
        self:push8(band(val, 0xff00) / 0x100)
        return self:push8(val % 0x100)
      end,
      pop16 = function(self, val)
        return bor(self:pop8(), lshift(self:pop8(), 8))
      end,
      setC = function(self, val)
        self.P = band(bor(self.P, val), 0xfe + val)
      end,
      setZ = function(self, val)
        self.P = band(bor(self.P, val), 0xfd + val)
      end,
      setI = function(self, val)
        self.P = band(bor(self.P, val), 0xfb + val)
      end,
      setD = function(self, val)
        self.P = band(bor(self.P, val), 0xf7 + val)
      end,
      setV = function(self, val)
        self.P = band(bor(self.P, val), 0xbf + val)
      end,
      setN = function(self, val)
        self.P = band(bor(self.P, val), 0x7f + val)
      end,
      updateV = function(self, op, p, m)
        return self:setV(band(bxor(op, p), band(bxor(op, m), 0x80)) / 2)
      end,
      updateZ = function(self, op)
        return self:setZ(2 - min(op, 1) * 2)
      end,
      updateN = function(self, op)
        return self:setN(band(op, 0x80))
      end,
      updateZN = function(self, op)
        local p = self.P
        if op == 0 then
          p = bor(p, 0x02) % 0x80
        else
          local b = band(op, 0x80)
          p = band(bor(band(p, 0xfd), b), 0x7f + b)
        end
        self.P = p
        return op
      end,
      testStatus = function(self, flag)
        return band(self.P, flag) == flag
      end,
      reset = function(self, addr)
        self.A, self.X, self.Y, self.SP, self.P = 0, 0, 0, 0xfd, 0x20
        self.PC = addr or self:read16(RESET)
        self:write8(STACK + 0xff, 0)
        self._port[0], self._port[1] = 0x2f, 0x37
        self._ir_pending = 0
        self._cycles = 0
        self._instCounter = 0
      end,
      irq = function(self)
        self._ir_pending = IRQ
      end,
      nmi = function(self)
        self._ir_pending = MRI
      end,
      _interrupt = function(self)
        local addr = self._ir_pending
        self._ir_pending = 0
        self:push16(self.PC)
        self:push8(band(self.P, 0xef))
        self:setI(0x04)
        self.PC = self:read16(addr)
        self._cycles = self._cycles + 7
      end,
      _nextInst = function(self)
        local b = self:read8(self.PC)
        local size, cycles, fn
        do
          local _obj_0 = OPCODES[b]
          size, cycles, fn = _obj_0[1], _obj_0[2], _obj_0[3]
        end
        self._cycles = self._cycles + cycles
        self._instCounter = self._instCounter + 1
        self.PC = self.PC + 1
        fn(self)
        self.PC = self.PC + (size - 1)
      end
    }
  })()
end
return function(mem, startAddr)
  return CPU(mem, startAddr)
end
