local ffi = require('ffi')
local bor, band, bxor, bnot, lshift, rshift
do
  local _obj_0 = require('bit')
  bor, band, bxor, bnot, lshift, rshift = _obj_0.bor, _obj_0.band, _obj_0.bxor, _obj_0.bnot, _obj_0.lshift, _obj_0.rshift
end
local RAM = {
  0x0000,
  0xffff
}
local CART = {
  0x8000,
  0x9fff
}
local BASIC = {
  0xa000,
  0xbfff
}
local CHAR = {
  0xd000,
  0xdfff
}
local IO = {
  0xd000,
  0xdfff
}
local KERNAL = {
  0xe000,
  0xffff
}
local Mem_cdef = [[	struct {
		uint8_t ram[0x10000];
		uint8_t cart[0x2000];
		uint8_t brom[0x2000];
		uint8_t crom[0x1000];
		uint8_t   io[0x1000]; // FIXME: only color ram 0xd800 - 0xdbff
		uint8_t krom[0x2000];
		uint8_t*   pt[0x100];
		uint8_t ioenabled;
	}
]]
return function(portRead, portWrite, ioRead, ioWrite)
  return ffi.metatype(Mem_cdef, {
    __new = function(self)
      local mem = ffi.new(self)
      do
        if CPU_TEST then
          if CPU_TEST then
            mem:switchRAM(0x0000, 0xffff)
          end
        else
          mem:_initROM()
          mem:switchRAM(RAM[1], BASIC[1] - 1)
          mem:switchBASIC()
          mem:switchRAM(BASIC[2] + 1, KERNAL[1] - 1)
          mem:switchIO()
          mem:switchKERNAL()
          mem:setIO(true)
        end
      end
      return mem
    end,
    __index = {
      load = function(self, data, addr)
        if addr == nil then
          addr = 0
        end
        assert(addr + #data <= 0x10000, "Data too large")
        return ffi.copy(self.ram + addr, ffi.cast('char*', data), #data)
      end,
      read8 = function(self, addr)
        return self:_read(addr)
      end,
      write8 = function(self, addr, value)
        return self:_write(addr, value)
      end,
      read16 = function(self, addr)
        return bor(self:_read(addr), lshift(self:_read(addr + 1), 8))
      end,
      write16 = function(self, addr, value)
        self:_write(addr, band(rshift(val, 8), 0xff))
        return self:_write(addr + 1, band(val, 0xff))
      end,
      setIO = function(self, val)
        self.ioenabled = val and 0xff or 0
      end,
      switchRAM = function(self, lo, hi)
        return self:_mapMemory(lo, hi, self.ram)
      end,
      switchKERNAL = function(self)
        return self:_mapMemory(KERNAL[1], KERNAL[2], self.krom)
      end,
      switchBASIC = function(self)
        return self:_mapMemory(BASIC[1], BASIC[2], self.brom)
      end,
      switchCHAR = function(self)
        return self:_mapMemory(CHAR[1], CHAR[2], self.crom)
      end,
      switchIO = function(self)
        return self:_mapMemory(IO[1], IO[2], self.io)
      end,
      switchCART = function(self)
        return self:_mapMemory(CART[1], CART[2], self.cart)
      end,
      _mapMemory = function(self, lo, hi, target)
        lo, hi = rshift(lo, 8), rshift(hi, 8)
        for i = lo, hi do
          self.pt[i] = target + lshift(i - lo, 8)
        end
      end,
      _read = function(self, addr)
        if addr > 1 then
          if self.ioenabled ~= 0 and band(addr, 0xf000) == 0xd000 then
            return ioRead(addr)
          end
          return self:_readRaw(addr)
        end
        return portRead(addr)
      end,
      _write = function(self, addr, val)
        if addr > 1 then
          if self.ioenabled ~= 0 and band(addr, 0xf000) == 0xd000 then
            return ioWrite(addr, val)
          end
          return self:_writeRaw(addr, val)
        end
        return portWrite(addr, val)
      end,
      _readRaw = function(self, addr)
        return self.pt[rshift(addr, 8)][band(addr, 0xff)]
      end,
      _writeRaw = function(self, addr, val)
        self.ram[addr] = val
      end,
      _initROM = function(self)
        do
          local kernal = love.filesystem.read('roms/kernal.rom')
          ffi.copy(self.krom, kernal, KERNAL[2] - KERNAL[1] + 1)
        end
        do
          local basic = love.filesystem.read('roms/basic.rom')
          ffi.copy(self.brom, basic, BASIC[2] - BASIC[1] + 1)
        end
        do
          local char = love.filesystem.read('roms/char.rom')
          return ffi.copy(self.crom, char, CHAR[2] - CHAR[1] + 1)
        end
      end
    }
  })()
end
